# Original code from https://beckmw.wordpress.com/2014/01/21/a-brief-foray-into-parallel-processing-with-r/
# Modified by Le Yan

# Workload: 
# Create 1,000 random samples, each with 1,000,000 observations from a standard normal distribution.
# Take a summary for each sample.

iters <- 1000

# Sequential
seqTime <- system.time(
for (i in 1:iters) {
  to.ls <- rnorm(1e6)
  to.ls <- summary(to.ls)
}
)

print("Sequential:")
print(seqTime)

# Multiple cores on the same node

library(doParallel)

# Detect the number of CPU cores.

# Find the number of cores
cores <- detectCores(logical=F)
print(paste("There are a total of",cores,"CPU cores on this computer."))

clusterSize <- c(2,4,8,16,32,64)
times <- vector("numeric",length(clusterSize))

for (i in 1:length(clusterSize)) {

  cl <- makeCluster(clusterSize[i])
  registerDoParallel(cl)

  t <- system.time(
  ls<-foreach(icount(iters)) %dopar% {
     
    to.ls<-rnorm(1e6)
    to.ls<-summary(to.ls)
     
    }
  )

  print(paste("Parallel with",clusterSize[i],"workers on one node:"))
  print(t)

  # Record the wall time.
  times[i] <- t["elapsed"]

  stopCluster(cl)

}

bm <- data.frame(worker=c(1,clusterSize), oneNodeTime=c(seqTime["elapsed"],times))

# Multiple nodes

hosts <- as.vector(unique(read.table(Sys.getenv("PBS_NODEFILE"),stringsAsFactors=F))[,1])
nh <- length(hosts)

if (nh != 2) {
  stop("The number of nodes must be 2!")
}

for (i in 1:length(clusterSize)) {

  cl <- makePSOCKcluster(rep(hosts , each = clusterSize[i]/nh))
  registerDoParallel(cl)

  t <- system.time(
  ls<-foreach(icount(iters)) %dopar% {
     
    to.ls<-rnorm(1e6)
    to.ls<-summary(to.ls)
     
    }
  )
 
  print(paste("Parallel with",clusterSize[i],"workers on",nh,"nodes:"))
  print(t)

  times[i] <- t["elapsed"]

  stopCluster(cl)

}

library(reshape)

bm[2:nrow(bm),"multiNodeTime"] <- times
bm[1,3] <- bm[1,2]
colnames(bm) <- c("Worker","On one node","On multiple nodes")
bm.long <- melt(bm,id="Worker",measure=c(2,3))

# Plot the times.

library(ggplot2)

png(filename="R_parallel_benchmarki_rnorm.png", width = 800, height = 600)

g <- ggplot(bm.long,aes(Worker, value, color=variable))+
  geom_point(size=3)+
  geom_line(size=2)+
  geom_text(data=subset(bm.long,variable=="On one node"),aes(label=Worker),vjust=4)+
  ggtitle('Wall clock time vs. number of workers for rnorm()')+
  xlab("Number of workers")+ylab("Wall clock time")+
  xlim(1,100)+ylim(1,200)+
  scale_y_log10(breaks=c(100,50,20,10,5))+
  scale_x_log10()+
  theme(plot.title = element_text(size=20),legend.title=element_blank(),text=element_text(size=16))

print(g)

dev.off()



